"""SCIM constants"""

from django_scim import constants as djs_constants


class SchemaURI(djs_constants.SchemaURI):
    BULK_REQUEST: str = "urn:ietf:params:scim:api:messages:2.0:BulkRequest"
    BULK_RESPONSE: str = "urn:ietf:params:scim:api:messages:2.0:BulkResponse"


SORT_MAPPING = {
    "id": "id",
    "userName": "username",
    "email": "email",
}

VALID_SORTS = SORT_MAPPING.keys()



from mitol.common.envs import get_int, get_string

MITOL_SCIM_REQUESTS_TIMEOUT_SECONDS = get_int(
    name="MITOL_SCIM_REQUESTS_TIMEOUT_SECONDS",
    default=60,
    description="Number of seconds to timeout requests to Keycloak",
)
MITOL_SCIM_KEYCLOAK_BATCH_SIZE = get_int(
    name="MITOL_SCIM_KEYCLOAK_BATCH_SIZE",
    default=250,
    description="Number of operations to send in a single batch request",
)
MITOL_SCIM_KEYCLOAK_SEARCH_BATCH_SIZE = get_int(
    name="MITOL_SCIM_KEYCLOAK_SEARCH_BATCH_SIZE",
    default=100,
    description="Number of operations to send in a single search request",
)

MITOL_SCIM_KEYCLOAK_BULK_OPERATIONS_COUNT = get_int(
    name="MITOL_SCIM_KEYCLOAK_BULK_OPERATIONS_COUNT",
    default=250,
    description="Number of operations to perform per bulk request",
    required=True,
)

MITOL_SCIM_KEYCLOAK_CLIENT_ID = get_string(
    name="MITOL_SCIM_KEYCLOAK_CLIENT_ID",
    description="The client id for the Keycloak service",
)
MITOL_SCIM_KEYCLOAK_CLIENT_SECRET = get_string(
    name="MITOL_SCIM_KEYCLOAK_CLIENT_SECRET",
    description="The client secret for the Keycloak service",
)

MITOL_SCIM_KEYCLOAK_BASE_URL = get_string(
    name="MITOL_SCIM_KEYCLOAK_BASE_URL",
    description="The base url for the upstream Keycloak service",
)

SCIM_SERVICE_PROVIDER = {
    "SCHEME": "https",
    # use default value,
    # this will be overridden by value returned by BASE_LOCATION_GETTER
    "NETLOC": "localhost",
    "AUTHENTICATION_SCHEMES": [
        {
            "type": "oauth2",
            "name": "OAuth 2",
            "description": "Oauth 2 implemented with bearer token",
            "specUri": "",
            "documentationUri": "",
        },
    ],
    "SERVICE_PROVIDER_CONFIG_MODEL": "mitol.scim.config.LearnSCIMServiceProviderConfig",
    "USER_ADAPTER": "mitol.scim.adapters.UserAdapter",
    "USER_MODEL_GETTER": "mitol.scim.adapters.get_user_model_for_scim",
    "USER_FILTER_PARSER": "mitol.scim.filters.UserFilterQuery",
    "GET_IS_AUTHENTICATED_PREDICATE": "mitol.scim.utils.is_authenticated_predicate",
}



import json
import logging
from typing import Optional, Union

from django.contrib.auth import get_user_model
from django.db import transaction
from django_scim.adapters import SCIMUser
from scim2_filter_parser.attr_paths import AttrPath

from mitol.scim.constants import SchemaURI

User = get_user_model()


logger = logging.getLogger(__name__)


def get_user_model_for_scim():
    """
    Get function for the django_scim library configuration (USER_MODEL_GETTER).

    Returns:
        model: User model.
    """
    return User


class UserAdapter(SCIMUser):
    """
    Custom adapter to extend django_scim library.
    """

    password_changed = False
    activity_changed = False

    resource_type = "User"

    id_field = "scim_id"

    ATTR_MAP = {
        ("active", None, None): "is_active",
        ("name", "givenName", None): "first_name",
        ("name", "familyName", None): "last_name",
        ("userName", None, None): "username",
    }

    IGNORED_PATHS = {
        ("schemas", None, None),
    }

    @property
    def is_new_user(self):
        """

        Returns:
            bool: True is the user does not currently exist,
            False if the user already exists.
        """
        return not bool(self.obj.id)

    @property
    def id(self):
        """
        Return the SCIM id
        """
        return self.obj.scim_id

    @property
    def emails(self):
        """
        Return the email of the user per the SCIM spec.
        """
        return [{"value": self.obj.email.lower(), "primary": True}]

    @property
    def display_name(self):
        """
        Return the displayName of the user per the SCIM spec.
        """
        return f"{self.obj.first_name} {self.obj.last_name}"

    @property
    def meta(self):
        """
        Return the meta object of the user per the SCIM spec.
        """
        return {
            "resourceType": self.resource_type,
            "created": self.obj.created_on.isoformat(timespec="milliseconds"),
            "lastModified": self.obj.updated_on.isoformat(timespec="milliseconds"),
            "location": self.location,
        }

    def to_dict(self):
        """
        Return a ``dict`` conforming to the SCIM User Schema,
        ready for conversion to a JSON object.
        """
        return {
            "id": self.id,
            "externalId": self.obj.scim_external_id,
            "schemas": [SchemaURI.USER],
            "userName": self.obj.username,
            "name": {
                "givenName": self.obj.first_name,
                "familyName": self.obj.last_name,
            },
            "displayName": self.display_name,
            "emails": self.emails,
            "active": self.obj.is_active,
            "groups": [],
            "meta": self.meta,
        }

    def from_dict(self, d):
        """
        Consume a ``dict`` conforming to the SCIM User Schema, updating the
        internal user object with data from the ``dict``.

        Please note, the user object is not saved within this method. To
        persist the changes made by this method, please call ``.save()`` on the
        adapter. Eg::

            scim_user.from_dict(d)
            scim_user.save()
        """
        self.parse_emails(d.get("emails"))

        self.obj.is_active = d.get("active", True)
        self.obj.username = d.get("userName")
        self.obj.first_name = d.get("name", {}).get("givenName", "")
        self.obj.last_name = d.get("name", {}).get("familyName", "")
        self.obj.scim_username = d.get("userName")
        self.obj.scim_external_id = d.get("externalId")
        self.obj.global_id = self.obj.scim_external_id or ""

    def _save_user(self):
        self.obj.save()

    def _save_related(self):
        pass

    def save(self):
        """
        Save instances of the Profile and User models.
        """
        with transaction.atomic():
            # user must be saved first due to FK Profile -> User
            self._save_user()
            self._save_related()
            logger.info("User saved. User id %i", self.obj.id)

    def delete(self):
        """
        Update User's is_active to False.
        """
        self.obj.is_active = False
        self.obj.save()
        logger.info("Deactivated user id %i", self.obj.id)

    def handle_add(
        self,
        path: Optional[AttrPath],
        value: Union[str, list, dict],
        operation: dict,  # noqa: ARG002
    ):
        """
        Handle add operations per:
        https://tools.ietf.org/html/rfc7644#section-3.5.2.1

        Args:
            path (AttrPath)
            value (Union[str, list, dict])
        """
        if path is None:
            return

        if path.first_path == ("externalId", None, None):
            self.obj.scim_external_id = value
            self.obj.save()

    def parse_scim_for_keycloak_payload(self, payload: str) -> dict:
        """
        Parse the payload sent from scim-for-keycloak and normalize it
        """
        result = {}

        for key, value in json.loads(payload).items():
            if key == "schema":
                continue

            if isinstance(value, dict):
                for nested_key, nested_value in value.items():
                    result[self.split_path(f"{key}.{nested_key}")] = nested_value
            else:
                result[key] = value

        return result

    def parse_path_and_values(
        self, path: Optional[str], value: Union[str, list, dict]
    ) -> list:
        """Parse the incoming value(s)"""
        if isinstance(value, str):
            # scim-for-keycloak sends this as a noncompliant JSON-encoded string
            if path is None:
                val = json.loads(value)
            else:
                msg = "Called with a non-null path and a str value"
                raise ValueError(msg)
        else:
            val = value

        results = []

        for attr_path, attr_value in val.items():
            if isinstance(attr_value, dict):
                # nested object, we want to recursively flatten it to `first.second`
                results.extend(self.parse_path_and_values(attr_path, attr_value))
            else:
                flattened_path = (
                    f"{path}.{attr_path}" if path is not None else attr_path
                )
                new_path = self.split_path(flattened_path)
                new_value = attr_value
                results.append((new_path, new_value))

        return results

    def _handle_resplace_nested_path(self, nested_path, nested_value):
        """Handle processing a nested path"""
        if nested_path.first_path in self.ATTR_MAP:
            setattr(self.obj, self.ATTR_MAP[nested_path.first_path], nested_value)
        elif nested_path.first_path == ("emails", None, None):
            self.parse_emails(nested_value)
        else:
            return False
        return True

    def handle_replace(
        self,
        path: Optional[AttrPath],
        value: Union[str, list, dict],
        operation: dict,  # noqa: ARG002
    ):
        """
        Handle the replace operations.

        All operations happen within an atomic transaction.
        """

        if not isinstance(value, dict):
            # Restructure for use in loop below.
            value = {path: value}

        for nested_path, nested_value in (value or {}).items():
            if (
                not self._handle_resplace_nested_path(nested_path, nested_value)
                and nested_path.first_path not in self.IGNORED_PATHS
            ):
                logger.debug(
                    "Ignoring SCIM update for path: %s", nested_path.first_path
                )

        self.save()



"""Sync users with SCIM to the remote"""

from django.core.management.base import BaseCommand


class Command(BaseCommand):
    """Sync users with SCIM to the remote"""

    help = "Sync users with SCIM to the remote"

    def add_arguments(self, parser) -> None:
        parser.add_argument(
            "--never-synced-only",
            action="store_true",
            default=False,
            help="Only sync users who have never been synced",
        )

    def handle(self, *args, **options):  # noqa: ARG002
        """Sync users with SCIM to the remote"""
        from mitol.scim import tasks

        never_synced_only = options["never_synced_only"]

        task = tasks.sync_all_users_to_scim_remote.delay(
            never_synced_only=never_synced_only
        )

        task.get()

        self.stdout.write(self.style.SUCCESS("Synced users to remote"))



"""
SCIM filter parsers
    + _tag_term_type(TermType.attr_name)

This module aims to compliantly parse SCIM filter queries per the spec:
https://datatracker.ietf.org/doc/html/rfc7644#section-3.4.2.2

Note that this implementation defines things slightly differently
because a naive implementation exactly matching the filter grammar will
result in hitting Python's recursion limit because the grammar defines
logical lists (AND/OR chains) as a recursive relationship.

This implementation avoids that by defining separately FilterExpr and
Filter. As a result of this, some definitions are collapsed and removed
(e.g. valFilter => FilterExpr).
"""

from enum import auto

try:
    from enum import StrEnum
except ImportError:
    from strenum import StrEnum

from pyparsing import (
    CaselessKeyword,
    Char,
    Combine,
    DelimitedList,
    FollowedBy,
    Forward,
    Group,
    Literal,
    Suppress,
    Tag,
    alphanums,
    alphas,
    common,
    dbl_quoted_string,
    nested_expr,
    one_of,
    remove_quotes,
    ungroup,
)


class TagName(StrEnum):
    """Tag names"""

    term_type = auto()
    value_type = auto()


class TermType(StrEnum):
    """Tag term type"""

    urn = auto()
    attr_name = auto()
    attr_path = auto()
    attr_expr = auto()
    value_path = auto()
    presence = auto()

    logical_op = auto()
    compare_op = auto()
    negation_op = auto()

    filter_expr = auto()
    filters = auto()


class ValueType(StrEnum):
    """Tag value_type"""

    boolean = auto()
    number = auto()
    string = auto()
    null = auto()


def _tag_term_type(term_type: TermType) -> Tag:
    return Tag(TagName.term_type.name, term_type)


def _tag_value_type(value_type: ValueType) -> Tag:
    return Tag(TagName.value_type.name, value_type)


NameChar = Char(alphanums + "_-")
AttrName = Combine(
    Char(alphas)
    + NameChar[...]
    # ensure we're not somehow parsing an URN
    + ~FollowedBy(":")
).set_results_name("attr_name") + _tag_term_type(TermType.attr_name)

# Example URN-qualifed attr:
# urn:ietf:params:scim:schemas:core:2.0:User:userName
# |--------------- URN --------------------|:| attr |
UrnAttr = Combine(
    Combine(
        Literal("urn:")
        + DelimitedList(
            # characters ONLY if followed by colon
            Char(alphanums + ".-_")[1, ...] + FollowedBy(":"),
            # separator
            Literal(":"),
            # combine everything back into a singular token
            combine=True,
        )[1, ...]
    ).set_results_name("urn")
    # separator between URN and attribute name
    + Literal(":")
    + AttrName
    + _tag_term_type(TermType.urn)
)


SubAttr = ungroup(Combine(Suppress(".") + AttrName)).set_results_name("sub_attr") ^ (
    Tag("sub_attr", None)
)

AttrPath = (
    (
        # match on UrnAttr first
        UrnAttr ^ AttrName
    )
    + SubAttr
    + _tag_term_type(TermType.attr_path)
)

ComparisonOperator = one_of(
    ["eq", "ne", "co", "sw", "ew", "gt", "lt", "ge", "le"],
    caseless=True,
    as_keyword=True,
).set_results_name("comparison_operator") + _tag_term_type(TermType.compare_op)

LogicalOperator = Group(
    one_of(["or", "and"], caseless=True).set_results_name("logical_operator")
    + _tag_term_type(TermType.logical_op)
)

NegationOperator = Group(
    (
        CaselessKeyword("not")
        + _tag_term_type(TermType.negation_op)
        + Tag("negated", True)  # noqa: FBT003
    )[..., 1]
    ^ Tag("negated", False)  # noqa: FBT003
)

ValueTrue = Literal("true").set_parse_action(lambda: True) + _tag_value_type(
    ValueType.boolean
)
ValueFalse = Literal("false").set_parse_action(lambda: False) + _tag_value_type(
    ValueType.boolean
)
ValueNull = Literal("null").set_parse_action(lambda: None) + _tag_value_type(
    ValueType.null
)
ValueNumber = (common.integer | common.fnumber) + _tag_value_type(ValueType.number)
ValueString = dbl_quoted_string.set_parse_action(remove_quotes) + _tag_value_type(
    ValueType.string
)

ComparisonValue = ungroup(
    ValueTrue | ValueFalse | ValueNull | ValueNumber | ValueString
).set_results_name("value")

AttrPresence = Group(
    AttrPath + Literal("pr").set_results_name("presence").set_parse_action(lambda: True)
) + _tag_term_type(TermType.presence)
AttrExpression = AttrPresence | Group(
    AttrPath + ComparisonOperator + ComparisonValue + _tag_term_type(TermType.attr_expr)
)

# these are forward references, so that we can have
# parsers circularly reference themselves
FilterExpr = Forward()
Filters = Forward()

ValuePath = Group(AttrPath + nested_expr("[", "]", Filters)).set_results_name(
    "value_path"
) + _tag_term_type(TermType.value_path)

FilterExpr <<= (
    AttrExpression | ValuePath | (NegationOperator + nested_expr("(", ")", Filters))
) + _tag_term_type(TermType.filter_expr)

Filters <<= (
    # comment to force it to wrap the below for operator precedence
    (FilterExpr + (LogicalOperator + FilterExpr)[...])
    + _tag_term_type(TermType.filters)
)



import logging

import celery
from django.contrib.auth import get_user_model
from django.db.models import Q

from mitol.common.utils.celery import get_celery_app
from mitol.common.utils.collections import chunks
from mitol.scim import api

User = get_user_model()
app = get_celery_app()


log = logging.getLogger()


@app.task(acks_late=True)
def sync_users_to_scim_remote_batch(*, user_ids: list[int]):
    """Sync a set of users to the scim remote"""
    users = User.objects.filter(id__in=user_ids).order_by("id")
    api.sync_users_to_scim_remote(users)


@app.task(bind=True, acks_late=True)
def sync_all_users_to_scim_remote(self, *, never_synced_only: bool = False):
    user_q = (
        User.objects.values_list("id", flat=True).filter(is_active=True).order_by("id")
    )

    if never_synced_only:
        user_q = user_q.filter(Q(global_id="") | Q(scim_external_id=None))

    msg = f"Syncing {user_q.count()} users to SCIM remote"
    log.info(msg)

    return self.replace(
        celery.group(
            sync_users_to_scim_remote_batch.si(user_ids=user_ids)
            for user_ids in chunks(
                user_q,
                chunk_size=1000,
            )
        )
    )



import operator
from collections.abc import Callable
from typing import Optional

from django.contrib.auth import get_user_model
from django.db.models import Model, Q
from pyparsing import ParseResults

from mitol.scim.parser import Filters, TermType


class FilterQuery:
    """Filters for users"""

    model_cls: type[Model]

    attr_map: dict[tuple[str, Optional[str]], tuple[str, ...]]

    related_selects: list[str] = []

    dj_op_mapping = {
        "eq": "exact",
        "ne": "exact",
        "gt": "gt",
        "ge": "gte",
        "lt": "lt",
        "le": "lte",
        "pr": "isnull",
        "co": "contains",
        "sw": "startswith",
        "ew": "endswith",
    }

    dj_negated_ops = ("ne", "pr")

    dj_op_mapping_per_field_overrides: dict[tuple[str, str | None], dict[str, str]] = {}

    @classmethod
    def _filter_expr(cls, parsed: ParseResults) -> Q:
        if parsed is None:
            msg = "Expected a filter, got: None"
            raise ValueError(msg)

        if parsed.term_type == TermType.attr_expr:
            return cls._attr_expr(parsed)

        msg = f"Unsupported term type: {parsed.term_type}"
        raise ValueError(msg)

    @classmethod
    def _attr_expr(cls, parsed: ParseResults) -> Q:
        scim_keys = (parsed.attr_name, parsed.sub_attr)

        field_op_overrides = cls.dj_op_mapping_per_field_overrides.get(scim_keys, {})

        op_name = parsed.comparison_operator.lower()
        dj_op = field_op_overrides.get(op_name, cls.dj_op_mapping[op_name])

        path_parts = list(
            filter(
                lambda part: part is not None,
                (
                    *cls.attr_map.get(scim_keys, scim_keys),
                    dj_op,
                ),
            )
        )
        path = "__".join(path_parts)

        q = Q(**{path: parsed.value})

        if parsed.comparison_operator in cls.dj_negated_ops:
            q = ~q

        return q

    @classmethod
    def _filters(cls, parsed: ParseResults) -> Q:
        parsed_iter = iter(parsed)
        q = cls._filter_expr(next(parsed_iter))

        try:
            while operator := cls._logical_op(next(parsed_iter)):
                filter_q = cls._filter_expr(next(parsed_iter))

                # combine the previous and next Q() objects using the bitwise operator
                q = operator(q, filter_q)
        except StopIteration:
            pass

        return q

    @classmethod
    def _logical_op(cls, parsed: ParseResults) -> Callable[[Q, Q], Q] | None:
        """Convert a defined operator to the corresponding bitwise operator"""
        if parsed is None:
            return None

        if parsed.logical_operator.lower() == "and":
            return operator.and_
        elif parsed.logical_operator.lower() == "or":
            return operator.or_
        else:
            msg = f"Unexpected operator: {parsed.operator}"
            raise ValueError(msg)

    @classmethod
    def search(cls, filter_query, request=None):  # noqa: ARG003
        """Create a search query"""
        parsed = Filters.parse_string(filter_query, parse_all=True)

        return cls.model_cls.objects.select_related(*cls.related_selects).filter(
            cls._filters(parsed)
        )


class UserFilterQuery(FilterQuery):
    """FilterQuery for User"""

    attr_map: dict[tuple[str, Optional[str]], tuple[str, ...]] = {
        ("userName", None): ("username",),
        ("emails", "value"): ("email",),
        ("active", None): ("is_active",),
        ("name", "givenName"): ("first_name",),
        ("name", "familyName"): ("last_name",),
    }

    related_selects = []

    dj_op_mapping_per_field_overrides = {
        ("emails", "value"): {
            "eq": "iexact",
            "ne": "iexact",
            "co": "contains",
            "sw": "startswith",
            "ew": "endswith",
        },
        ("emails", None): {
            "eq": "iexact",
            "ne": "iexact",
            "co": "contains",
            "sw": "startswith",
            "ew": "endswith",
        },
    }

    model_cls = get_user_model()



"""SCIM view customizations"""

import json
import logging
from http import HTTPStatus
from urllib.parse import urljoin, urlparse

from django.http import HttpResponse
from django.urls import Resolver404, resolve, reverse
from django_scim import constants as djs_constants
from django_scim import exceptions
from django_scim import views as djs_views
from django_scim.utils import get_base_scim_location_getter

from mitol.scim import constants
from mitol.scim.requests import InMemoryHttpRequest

log = logging.getLogger()


class BulkView(djs_views.SCIMView):
    http_method_names = ["post"]

    def post(self, request, *args, **kwargs):  # noqa: ARG002
        body = self.load_body(request.body)

        if body.get("schemas") != [constants.SchemaURI.BULK_REQUEST]:
            msg = "Invalid schema uri. Must be SearchRequest."
            raise exceptions.BadRequestError(msg)

        fail_on_errors = body.get("failOnErrors", None)

        if fail_on_errors is not None and not isinstance(fail_on_errors, int):
            msg = "Invalid failOnErrors. Must be an integer."
            raise exceptions.BaseRequestError(msg)

        operations = body.get("Operations")

        results = self._attempt_operations(request, operations, fail_on_errors)

        response = {
            "schemas": [constants.SchemaURI.BULK_RESPONSE],
            "Operations": results,
        }

        content = json.dumps(response)

        return HttpResponse(
            content=content,
            content_type=djs_constants.SCIM_CONTENT_TYPE,
            status=HTTPStatus.OK,
        )

    def _attempt_operations(self, request, operations, fail_on_errors):
        """Attempt to run the operations that were passed"""
        responses = []
        num_errors = 0

        for operation in operations:
            # per-spec,if we've hit the error threshold stop processing and return
            if fail_on_errors is not None and num_errors >= fail_on_errors:
                break

            op_response = self._attempt_operation(request, operation)

            # if the operation returned a non-2xx status code, record it as a failure
            if int(op_response.get("status")) >= HTTPStatus.MULTIPLE_CHOICES:
                num_errors += 1

            responses.append(op_response)

        return responses

    def _attempt_operation(self, bulk_request, operation):
        """Attempt an operation as part of a bulk request"""

        method = operation.get("method")
        bulk_id = operation.get("bulkId")
        path = operation.get("path")
        data = operation.get("data")

        try:
            url_match = resolve(path, urlconf="django_scim.urls")
        except Resolver404:
            return self._operation_error(
                bulk_id,
                HTTPStatus.NOT_IMPLEMENTED,
                "Endpoint is not supported for /Bulk",
            )

        # this is an ephemeral request not tied to the real request directly
        op_request = InMemoryHttpRequest.from_request(
            bulk_request, path, method, json.dumps(data).encode(djs_constants.ENCODING)
        )

        op_response = url_match.func(op_request, *url_match.args, **url_match.kwargs)
        result = {
            "method": method,
            "bulkId": bulk_id,
            "status": str(op_response.status_code),
        }

        location = None

        if op_response.status_code >= HTTPStatus.BAD_REQUEST and op_response.content:
            result["response"] = json.loads(op_response.content.decode("utf-8"))

        location = op_response.headers.get("Location", None)

        if location is not None:
            result["location"] = location
            # this is a custom field that the scim-for-keycloak plugin requires
            try:
                path = urlparse(location).path
                location_match = resolve(path)
                # this URL will be something like /scim/v2/Users/12345
                # resolving it gives the uuid
                result["id"] = location_match.kwargs["uuid"]
            except Resolver404:
                log.exception("Unable to resolve resource url: %s", location)

        return result

    def _operation_error(self, method, bulk_id, status_code, detail):
        """Return a failure response"""
        status_code = str(status_code)
        return {
            "method": method,
            "status": status_code,
            "bulkId": bulk_id,
            "response": {
                "schemas": [djs_constants.SchemaURI.ERROR],
                "status": status_code,
                "detail": detail,
            },
        }


class SearchView(djs_views.UserSearchView):
    """
    View for /.search endpoint
    """

    def post(self, request, *args, **kwargs):  # noqa: ARG002
        body = self.load_body(request.body)
        if body.get("schemas") != [djs_constants.SchemaURI.SERACH_REQUEST]:
            msg = "Invalid schema uri. Must be SearchRequest."
            raise exceptions.BadRequestError(msg)

        # cast to ints because scim-for-keycloak sends strings
        start = int(body.get("startIndex", 1))
        count = int(body.get("count", 50))
        sort_by = body.get("sortBy", "id")
        sort_order = body.get("sortOrder", "ascending")
        query = body.get("filter", None)

        if sort_by not in constants.VALID_SORTS:
            msg = f"Sorting only supports: {', '.join(constants.VALID_SORTS)}"
            raise exceptions.BadRequestError(msg)
        else:
            sort_by = constants.SORT_MAPPING[sort_by]

        if sort_order not in ("ascending", "descending"):
            msg = "Sorting only supports ascending or descending"
            raise exceptions.BadRequestError(msg)

        if not query:
            msg = "No filter query specified"
            raise exceptions.BadRequestError(msg)

        try:
            qs = self.__class__.parser_getter().search(query, request)
        except ValueError as e:
            msg = "Invalid filter/search query: " + str(e)
            raise exceptions.BadRequestError(msg) from e

        qs = qs.order_by(sort_by)

        if sort_order == "descending":
            qs = qs.reverse()

        response = self._build_response(request, qs, start, count)

        path = reverse(self.scim_adapter.url_name)
        url = urljoin(get_base_scim_location_getter()(request=request), path).rstrip(
            "/"
        )
        response["Location"] = url + "/.search"
        return response



import copy
from urllib.parse import urlparse

from django.conf import settings
from django.http import HttpRequest
from django_scim import constants as djs_constants


class InMemoryHttpRequest(HttpRequest):
    """
    A spoofed HttpRequest that only exists in-memory.
    It does not implement all features of HttpRequest and is only used
    for the bulk SCIM operations here so we can reuse view implementations.
    """

    @classmethod
    def from_request(cls, request: HttpRequest, path: str, method: str, body: str):
        return cls(request.META, path, method, body)

    @classmethod
    def stub(cls, *, path: str = "/", method: str = "GET", body: str = ""):
        parsed = urlparse(settings.SITE_BASE_URL)
        return cls(
            {
                "SERVER_HOST": parsed.hostname,
                "SERVER_PORT": parsed.port,
            },
            path,
            method,
            body,
        )

    def __init__(self, meta: dict[str, str], path: str, method: str, body: str):
        super().__init__()

        self.META = copy.deepcopy(
            {
                key: value
                for key, value in meta.items()
                if isinstance(value, str) and not key.startswith(("wsgi", "uwsgi"))
            }
        )
        self.path = path
        self.method = method
        self.content_type = djs_constants.SCIM_CONTENT_TYPE

        # normally HttpRequest would read this in, but we already have the value
        self._body = body



import http
import logging
from collections.abc import Generator
from dataclasses import dataclass
from typing import Any
from urllib.parse import urlsplit

import requests
from django.conf import settings
from django.contrib.auth import get_user_model
from django_scim.adapters import SCIMUser
from django_scim.utils import get_user_adapter
from more_itertools import chunked, first, partition
from oauthlib.oauth2 import BackendApplicationClient
from requests_oauthlib import OAuth2Session

from mitol.scim.constants import SchemaURI
from mitol.scim.requests import InMemoryHttpRequest

User = get_user_model()
UserAdapter: type[SCIMUser] = get_user_adapter()


log = logging.getLogger()


@dataclass()
class UserState:
    user: "User"
    external_id: str


@dataclass()
class UserOperation:
    user: "User"
    bulk_id: str
    operation: dict[str, Any]


StateOrOperation = UserState | UserOperation
StateOrOperationGenerator = Generator[StateOrOperation, Any, Any]
StateGenerator = Generator[UserState, Any, Any]


def realm_api_url(path: str) -> str:
    base_url: str = settings.MITOL_SCIM_KEYCLOAK_BASE_URL
    base_url = base_url.rstrip("/")
    path = path.lstrip("/")
    return f"{base_url}/{path}"


def scim_api_url(path: str) -> str:
    path = path.lstrip("/")
    return realm_api_url(f"/scim/v2/{path}")


def oidc_discovery_url():
    return realm_api_url("/.well-known/openid-configuration")


def get_openid_configuration():
    response = requests.get(
        oidc_discovery_url(), timeout=settings.MITOL_SCIM_REQUESTS_TIMEOUT_SECONDS
    )
    response.raise_for_status()
    return response.json()


def get_session() -> OAuth2Session:
    openid_config = get_openid_configuration()
    client_id = settings.MITOL_SCIM_KEYCLOAK_CLIENT_ID
    client_secret = settings.MITOL_SCIM_KEYCLOAK_CLIENT_SECRET

    token_url = openid_config["token_endpoint"]
    client = BackendApplicationClient(client_id=client_id)
    oauth = OAuth2Session(client=client)
    token = oauth.fetch_token(
        token_url=token_url,
        client_id=client_id,
        client_secret=client_secret,
    )

    return OAuth2Session(token=token)


def sync_users_to_scim_remote(users: list["User"]):
    with get_session() as session:
        found_users = _user_search_by_email(session, users)
        state_or_operations = _get_sync_operations(users, found_users)
        states = _perform_sync_operations(session, state_or_operations)
        _update_users(states)


def _user_search_by_email(
    session: OAuth2Session, users: list["User"]
) -> StateOrOperationGenerator:
    for users_batch in chunked(users, settings.MITOL_SCIM_KEYCLOAK_SEARCH_BATCH_SIZE):
        yield from _user_search_by_email_batch(session, users_batch)


def _user_search_by_email_batch(
    session: OAuth2Session, users: list["User"]
) -> StateOrOperationGenerator:
    """Perform a search for a set of users by email"""
    users_by_email = {user.email.lower(): user for user in users}

    payload = {
        "schemas": [SchemaURI.SERACH_REQUEST],  # typo in upstream lib
        "filter": " OR ".join(
            [f'emails.value EQ "{email}"' for email in users_by_email]
        ),
    }

    start_index = 1
    while True:
        resp = session.post(
            scim_api_url("/Users/.search"),
            json={
                **payload,
                "startIndex": start_index,
            },
            timeout=settings.MITOL_SCIM_REQUESTS_TIMEOUT_SECONDS,
        )

        if resp.status_code != http.HTTPStatus.OK:
            log.error("Error response: %s", resp.json())

        resp.raise_for_status()

        data = resp.json()

        resources = data.get("Resources", [])

        for resource in resources:
            email = first(
                [
                    email["value"].lower()
                    for email in resource.get("emails", [])
                    if email.get("primary", False)
                ]
            )
            if email is None:
                log.error("Unexpected user result with no email")
            elif email not in users_by_email:
                log.error(
                    "Received an email in search results that does not match: %s", email
                )
            else:
                yield UserState(users_by_email[email], resource["id"])

        if not resources:
            break

        start_index += len(resources)


def _parse_external_id_from_location(location: str) -> str:
    """Get the external id by taking the last part of the url path"""
    path = urlsplit(location).path
    return path.split("/")[-1]


def _get_sync_operations(
    users: list["User"], found_users: StateOrOperationGenerator
) -> StateOrOperationGenerator:
    """Generate the operations we need to perform"""
    missing_users = set(users)

    for user_resource in found_users:
        user = user_resource.user

        if user in missing_users:
            missing_users.remove(user)

        yield user_resource

    for user in missing_users:
        bulk_id = str(user.id)
        adapter = UserAdapter(
            # The request is only necessary because `SCIMUser.location` accesses
            # `SCIMUser.request`, whch raises an exception if the request is
            # `None` (yay side-effecting properties). However, the default
            # location getter doesn't even use the request. We create an in-memory
            # request that would be correct if it's ever used just in case.
            user,
            InMemoryHttpRequest.stub(),
        )
        yield UserOperation(
            user,
            bulk_id,
            {
                "method": "POST",
                "path": "/Users",
                "bulkId": bulk_id,
                "data": {
                    **adapter.to_dict(),
                    "emailVerified": True,
                },
            },
        )


def _perform_sync_operations(
    session: OAuth2Session,
    generator: StateOrOperationGenerator,
) -> StateGenerator:
    user_states, sync_operations = partition(
        lambda so: isinstance(so, UserOperation), generator
    )

    yield from user_states

    for chunk in chunked(
        sync_operations,
        settings.MITOL_SCIM_KEYCLOAK_BULK_OPERATIONS_COUNT,
    ):
        operations = []
        users_by_bulk_id: dict[int, User] = {}

        for state_or_operation in chunk:
            operations.append(state_or_operation.operation)
            users_by_bulk_id[state_or_operation.bulk_id] = state_or_operation.user

        if len(operations) == 0:
            # skip to the next chunk because this one had no operations
            # NOTE: this is not an indication we're at the end of chunks
            continue

        response = session.post(
            scim_api_url("/Users/Bulk"),
            json={
                "schemas": [SchemaURI.BULK_REQUEST],
                "Operations": operations,
            },
            timeout=settings.MITOL_SCIM_REQUESTS_TIMEOUT_SECONDS,
        )

        if response.status_code != http.HTTPStatus.OK:
            log.error("Error response: %s", response.json())

        response.raise_for_status()

        data = response.json()

        for operation in data["Operations"]:
            bulk_id = operation["bulkId"]
            user = users_by_bulk_id[bulk_id]

            if int(operation["status"]) != http.HTTPStatus.CREATED:
                log.error(
                    "Unable to perform operation for user: %s, response: %s",
                    str(user),
                    str(operation),
                )
                continue

            location = operation["location"]
            external_id = _parse_external_id_from_location(location)

            yield UserState(user, external_id)


def _update_users(states: StateGenerator):
    """Update the users to store the scim ids"""
    for batch in chunked(states, settings.MITOL_SCIM_KEYCLOAK_BULK_OPERATIONS_COUNT):
        updates = []
        for state in batch:
            user = state.user
            user.scim_id = str(user.id)  # normally done in User.save()
            user.scim_external_id = state.external_id
            user.global_id = state.external_id
            updates.append(user)

        User.objects.bulk_update(updates, ["global_id", "scim_id", "scim_external_id"])



import os

from django.apps import AppConfig


class ScimApp(AppConfig):
    name = "mitol.scim"
    label = "scim"
    verbose_name = "SCIM"

    required_settings = []

    default_auto_field = "django.db.models.BigAutoField"

    # necessary because this is a namespaced app
    path = os.path.dirname(os.path.abspath(__file__))  # noqa: PTH100, PTH120

    def ready(self) -> None:
        pass



"""URL configurations for SCIM"""

from django.urls import include, re_path

from mitol.scim import views

ol_scim_urls = (
    [
        re_path(r"^Bulk$", views.BulkView.as_view(), name="bulk"),
        re_path(r"^Users/\.search$", views.SearchView.as_view(), name="users-search"),
    ],
    "ol-scim",
)

urlpatterns = [
    re_path(r"^scim/v2/", include(ol_scim_urls)),
    re_path(r"^scim/v2/", include("django_scim.urls", namespace="scim")),
]



from django_scim.models import SCIMServiceProviderConfig


class LearnSCIMServiceProviderConfig(SCIMServiceProviderConfig):
    """Custom provider config"""

    def to_dict(self):
        result = super().to_dict()

        result["bulk"]["supported"] = True
        result["filter"]["supported"] = True

        return result
